@@ -1,10 +1,45 @@
 use no_std_net::{SocketAddr, ToSocketAddrs};
 use tinyvec::ArrayVec;
 
-/// Something that is associated with some network socket
+/// Data that came from a network socket
 #[derive(Debug, Clone, Copy)]
 pub struct Addressed<T>(pub T, pub SocketAddr);
 
+impl<T> Addressed<T> {
+  /// Borrow the contents of this Addressed
+  pub fn as_ref(&self) -> Addressed<&T> {
+    Addressed(self.data(), self.addr())
+  }
+
+  /// Discard the socket and get the data in this Addressed
+  pub fn unwrap(self) -> T {
+    self.0
+  }
+
+  /// Map the data contained in this Addressed
+  pub fn map<R>(self, f: impl FnOnce(T) -> R) -> Addressed<R> {
+    Addressed(f(self.0), self.1)
+  }
+
+  /// Map the data contained in this Addressed (with a copy of the address)
+  pub fn map_with_addr<R>(self, f: impl FnOnce(T, SocketAddr) -> R) -> Addressed<R> {
+    Addressed(f(self.0, self.1), self.1)
+  }
+
+  /// Borrow the contents of the addressed item
+  pub fn data(&self) -> &T {
+    &self.0
+  }
+
+  /// Copy the socket address for the data
+  pub fn addr(&self) -> SocketAddr {
+    self.1
+  }
+}
+
+/// TODO
+pub type Dgram = ArrayVec<[u8; 1152]>;
+
 /// A CoAP network socket
 ///
 /// This mirrors the Udp socket traits in embedded-nal, but allows us to implement them for foreign types (like `std::net::UdpSocket`).
@@ -25,12 +60,12 @@ pub trait Socket {
   fn recv(&self, buffer: &mut [u8]) -> nb::Result<(usize, SocketAddr), Self::Error>;
 
   /// Poll the socket for a datagram
-  fn poll(&self) -> Result<Option<(ArrayVec<[u8; 1152]>, SocketAddr)>, Self::Error> {
+  fn poll(&self) -> Result<Option<Addressed<Dgram>>, Self::Error> {
     let mut buf = [0u8; 1152];
     let recvd = self.recv(&mut buf);
 
     match recvd {
-      | Ok((n, addr)) => Ok(Some((buf.into_iter().take(n).collect(), addr))),
+      | Ok((n, addr)) => Ok(Some(Addressed(buf.into_iter().take(n).collect(), addr))),
       | Err(nb::Error::WouldBlock) => Ok(None),
       | Err(nb::Error::Other(e)) => Err(e),
     }
